import importlib
import unittest

from acq.django import discover
from django.conf import settings
from unittest import mock

settings.configure(AUTOLOAD_PACKAGES=[])

MOCK_INSTALLED_APPS = [
    'django.contrib.flatpages',
    'django.contrib.admin',
]


class DiscoveryTestCase(unittest.TestCase):
    class FancyClass:
        pass

    @mock.patch('acq.discovery.discover')
    def test_discover_imports_from_installed_apps(self, base_discover):
        with mock.patch('django.conf.settings.INSTALLED_APPS',
                        MOCK_INSTALLED_APPS):
            discover('views')

        base_discover.assert_called_once_with(
            'views',
            package_names=MOCK_INSTALLED_APPS,
            types=None,
        )

    @mock.patch('acq.discovery.discover')
    def test_discover_imports_from_provided_setting(self, base_discover):
        with mock.patch('django.conf.settings.AUTOLOAD_PACKAGES',
                        MOCK_INSTALLED_APPS):
            discover('views', packages_setting_name='AUTOLOAD_PACKAGES')

        base_discover.assert_called_once_with(
            'views',
            package_names=MOCK_INSTALLED_APPS,
            types=None,
        )

    @mock.patch('acq.discovery.discover')
    def test_discover_fails_if_setting_does_not_exist(self, base_discover):
        expected_msg = (
            "module 'django.conf.global_settings' has no "
            "attribute 'NON_EXISTANT_SETTING'"
        )

        with self.assertRaises(AttributeError, msg=expected_msg):
            discover(packages_setting_name='NON_EXISTANT_SETTING')

    @mock.patch('importlib.import_module')
    def test_discover_gets_instance_via_types(self, import_module):
        fake_module = mock.MagicMock()

        import_module.return_value = fake_module
        fake_module.FancyClass = self.FancyClass

        types = discover('example1', types=(self.FancyClass,))
        fancy_classes = list(types[self.FancyClass])

        self.assertEqual(len(fancy_classes), 1)
        self.assertIs(fancy_classes[0], self.FancyClass)



import types
import typing

from django.conf import settings

from acq import discovery

DEFAULT_MODULES_SETTING_NAME = 'INSTALLED_APPS'

T = typing.TypeVar("T", bound=typing.Type[object])


@typing.overload
def discover(
    *module_names: str,
    packages_setting_name: str,
    types: typing.Literal[None] = None,
) -> typing.List[types.ModuleType]:
    """ called with no types provided """


@typing.overload
def discover(
    *module_names: str,
    packages_setting_name: str,
    types: typing.Tuple[T],
) -> typing.Mapping[T, typing.Sequence[T]]:
    """ called with types provided """


def discover(
    *module_names: str,
    packages_setting_name: str = DEFAULT_MODULES_SETTING_NAME,
    types: None | typing.Tuple[T] = None,
) -> typing.List[types.ModuleType] | typing.Mapping[T, typing.Sequence[T]]:
    """ Discover modules from packages listed in INSTALLED_APPS

    Provides a function to detect modules within INSTALLED_APPS or a given
    setting name. When provided, the setting with a name matching
    packages_setting_name will be searched instead of INSTALLED_APPS.

    TODO: Add support for AppConfig

    """

    print(settings)
    packages = getattr(settings, packages_setting_name)

    return discovery.discover(
        *module_names,
        package_names=packages,
        types=types,
    )



import collections
import importlib
import inspect
import types
import typing

MODULE_NAME_SEPARATOR = '.'

T = typing.TypeVar("T", bound=typing.Type[object])


def import_module_safe(module_name: str) -> typing.Optional[types.ModuleType]:
    """ Wrapper for importlib.import_module that ignores ImportError

    """

    try:
        return importlib.import_module(module_name)
    except ImportError:
        return None


def get_module_name(*parts: str) -> str:
    return MODULE_NAME_SEPARATOR.join(parts)


def get_modules(
    *module_names: str,
    package_names: typing.Sequence[str] = [],
) -> typing.List[types.ModuleType]:
    if not package_names:
        return list(filter(None, map(import_module_safe, module_names)))

    if not module_names:
        return get_modules(*package_names)

    all_module_names = []

    for package_name in package_names:
        for module_name in module_names:
            all_module_names.append(get_module_name(package_name, module_name))

    return get_modules(*all_module_names)


def get_module_objects(
    modules: typing.List[types.ModuleType],
    types: typing.Tuple[T],
) -> typing.Mapping[T, typing.Sequence[T]]:
    results = collections.defaultdict(list)

    for module in modules:
        for attribute_name in dir(module):
            for object_type in types:
                module_object = getattr(module, attribute_name)
                is_class = inspect.isclass(module_object)

                if attribute_name == 'FancySubClass':
                    print(
                        module_object,
                        issubclass(module_object, object_type),
                    )

                if module_object is object_type:
                    results[object_type].append(module_object)
                elif is_class and issubclass(module_object, object_type):
                    results[object_type].append(module_object)
                elif isinstance(module_object, object_type):
                    results[object_type].append(module_object)

    return results


@typing.overload
def discover(
    *module_names: str,
    package_names: typing.Sequence[str] = [],
    types: typing.Literal[None] = None,
) -> typing.List[types.ModuleType]:
    """ called with no types provided """


@typing.overload
def discover(
    *module_names: str,
    package_names: typing.Sequence[str] = [],
    types: typing.Tuple[T],
) -> typing.Mapping[T, typing.Sequence[T]]:
    """ called with types provided """


def discover(
    *module_names: str,
    package_names: typing.Sequence[str] = [],
    types: None | typing.Tuple[T] = None,
) -> typing.List[types.ModuleType] | typing.Mapping[T, typing.Sequence[T]]:
    """ Allows discovery of modules, packages, or objects within them """

    modules = get_modules(*module_names, package_names=package_names)

    if types is None:
        return modules

    return get_module_objects(modules, types)



import acq
import unittest


class PublishingTestCase(unittest.TestCase):
    def test_name_is_correct(self):
        assert acq.name() == 'acq'

    def test_version_is_correct(self):
        assert acq.version_string() == '.'.join(
            map(
                str, [
                    acq.__version__[0],
                    acq.__version__[1],
                    acq.__version__[2],
                ]
            )
        )

    def test_version_string_with_release(self):
        new_version = (0, 0, 1, 'a0')
        assert acq.version_string(new_version) == '0.0.1-a0'

    def test_long_description_is_non_empty(self):
        assert len(acq.long_description().strip())

    def test_short_description_is_non_empty(self):
        assert len(acq.short_description().strip())



import unittest

from acq.discovery import discover

from unittest import mock

NOT_A_MODULE = 'Not A Module'
NOT_ANOTHER_MODULE = 'Not Another Module'
NOT_YET_ANOTHER_MODULE = 'Not Yet Another Module'


class DiscoveryTestCase(unittest.TestCase):
    class FancyClass:
        pass

    class FancySubClass(FancyClass):
        pass

    @mock.patch('importlib.import_module')
    def test_discover_imports_expected_module(self, import_module):
        import_module.return_value = NOT_A_MODULE

        modules = discover('example1')

        import_module.assert_called_once_with('example1')
        assert modules[0] == NOT_A_MODULE

    @mock.patch('importlib.import_module')
    def test_discover_imports_multiple_modules(self, import_module):
        import_module.side_effect = [
            NOT_A_MODULE,
            NOT_ANOTHER_MODULE,
            NOT_YET_ANOTHER_MODULE,
        ]

        modules = discover('example1', 'example2', 'example3')

        import_module.assert_has_calls(
            [
                mock.call('example1'),
                mock.call('example2'),
                mock.call('example3'),
            ]
        )

        assert modules[0] == NOT_A_MODULE
        assert modules[1] == NOT_ANOTHER_MODULE

    @mock.patch('importlib.import_module')
    def test_discover_called_with_no_modules(self, import_module):
        assert discover() == []
        assert import_module.call_count == 0

    @mock.patch('importlib.import_module')
    def test_discover_ignores_non_existant_modules(self, import_module):
        import_module.side_effect = [
            NOT_A_MODULE,
            ImportError('That is not a module.'),
            NOT_ANOTHER_MODULE,
        ]
        modules = discover('example1', 'example2', 'example3')
        assert len(modules) == 2

    @mock.patch('importlib.import_module')
    def test_discover_called_with_package_names(self, import_module):
        import_module.side_effect = [
            NOT_A_MODULE,
            NOT_ANOTHER_MODULE,
        ]
        modules = discover(package_names=['example1', 'example2'])

        assert len(modules) == 2
        assert modules[0] == NOT_A_MODULE
        assert modules[1] == NOT_ANOTHER_MODULE

    @mock.patch('importlib.import_module')
    def test_discover_called_with_package_names_and_module_names(
        self, import_module
    ):
        discover(
            'registration',
            'identification',
            package_names=['example1', 'example2'],
        )

        assert import_module.call_count == 4
        import_module.assert_has_calls(
            [
                mock.call('example1.registration'),
                mock.call('example1.identification'),
                mock.call('example2.registration'),
                mock.call('example2.identification'),
            ],
            any_order=True,
        )

    @mock.patch('importlib.import_module')
    def test_discover_gets_matching_instance_via_types(self, import_module):
        fake_module = mock.MagicMock()
        import_module.return_value = fake_module

        fake_module.FancyClass = self.FancyClass

        types = discover('example1', types=(self.FancyClass,))
        fancy_classes = list(types[self.FancyClass])

        self.assertEqual(len(fancy_classes), 1)
        self.assertIs(fancy_classes[0], self.FancyClass)

    @mock.patch('importlib.import_module')
    def test_discover_gets_class_via_types(self, import_module):
        fake_module = mock.MagicMock()
        import_module.return_value = fake_module

        fake_module.FancySubClass = self.FancySubClass

        types = discover('example1', types=(self.FancyClass,))
        fancy_classes = list(types[self.FancyClass])

        self.assertEqual(len(fancy_classes), 1)
        self.assertIs(fancy_classes[0], self.FancySubClass)

    @mock.patch('importlib.import_module')
    def test_discover_gets_instance_via_types(self, import_module):
        fake_module = mock.MagicMock()
        instance = self.FancySubClass()

        import_module.return_value = fake_module
        fake_module.fancy_sub_instance = instance

        types = discover('example1', types=(self.FancyClass,))
        fancy_classes = list(types[self.FancyClass])

        self.assertEqual(len(fancy_classes), 1)
        self.assertIs(fancy_classes[0], instance)



import acq
import setuptools

# setup.cfg is a new-ish standard, so we need to check this for now
if int(setuptools.__version__.split('.', 1)[0]) < 38:
    raise EnvironmentError(
        'Please upgrade setuptools. This package uses setup.cfg, which requires '
        'setuptools version 38 or higher. If you use pip, for instance, you can '
        'upgrade easily with ` pip install -U setuptools `'
    )

setuptools.setup(
    description=acq.short_description(),
    long_description=acq.long_description(),
    name=acq.name(),
    version=acq.version_string(),
)



""" automatic acquisition of modules from packages

Provides a simple and convenient way to automatically search packages for
modules and import them without needing project files to manually import them,
which allows for automatic configuration and other useful mechanics within
projects.

"""

__all__ = (
    '__version__',
    'version_string',
    'name',
    'long_description',
    'short_description',
    'discover',
)

__version__ = (0, 2, 0)

import typing

from .discovery import discover

VersionType = typing.Tuple[int, int, int] | typing.Tuple[int, int, int, str]


def version_string(version: VersionType = __version__) -> str:
    """ Provides the version string for this project """

    version_string = '.'.join(map(str, version[:3]))

    if len(version) <= 3:
        return version_string

    return version_string + '-' + version[3]


def name() -> str:
    """ Provides the package name for this project """

    return __name__


def long_description(content: str = ' '.join(__doc__.split('\n')[1:])) -> str:
    """ Provides a human-readable description for this project """

    return content


def short_description() -> str:
    """ Provides a summarized project desription """

    return long_description(content=__doc__.split('\n', 1)[0])
